/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	NJN/RBD
 *
 *	$Id: lamdtype.c,v 6.1 96/11/23 22:52:59 nevin Rel $
 *
 *	Function:	- miscellaneous datatype functions
 */

#include <errno.h>
#include <stdlib.h>

#include <mpi.h>
#include <mpisys.h>
#include <mpitrace.h>
#include <rpisys.h>

/*
 * global functions
 */
int			lam_type_free();
int			lam_dtsndrcv();
void			lam_dtcpy();
void			lam_dtblock();
void			lam_dtalign();
int			lam_dtbuffer();

/*
 * external functions
 */
extern void		lam_tr_dtypefree();

/*
 * local functions
 */
static void		cpy_hvector();
static void		cpy_hindexed();
static void		cpy_struct();


/*
 *	lam_type_free
 *
 *	Function:	- really free a datatype
 *	Accepts:	- datatype
 *	Returns:	- 0 or LAMERROR
 */
int
lam_type_free(dt)

MPI_Datatype		dt;

{
	int		err;			/* error code */
	int		i;
/*
 * Recursively free all datatype children.
 */
	switch(dt->dt_format) {

	case LAM_DTCONTIG:
	case LAM_DTVECTOR:
	case LAM_DTHVECTOR:
	case LAM_DTINDEXED:
	case LAM_DTHINDEXED:

	if (!(dt->dt_dtype->dt_flags & LAM_PREDEF)) {
		err = MPI_Type_free(&(dt->dt_dtype));
		if (err != MPI_SUCCESS) return(err);
	}

	if (dt->dt_format == LAM_DTHINDEXED || dt->dt_format == LAM_DTINDEXED) {
		if (dt->dt_disps) free((char *) dt->dt_disps);
	}

	break;

	case LAM_DTSTRUCT:

	for (i = 0; i < dt->dt_count; ++i) {
		if (!(dt->dt_dtypes[i]->dt_flags & LAM_PREDEF)) {
			err = MPI_Type_free(&(dt->dt_dtypes[i]));
			if (err != MPI_SUCCESS) return(err);
		}
	}

	if (dt->dt_dtypes) free((char *) dt->dt_dtypes);
	break;

	case LAM_DTBASIC:
		return(lam_mkerr(MPI_ERR_TYPE, 0));

	default:
		return(lam_mkerr(MPI_ERR_ARG, 0));
	}
/*
 * Free the top datatype.
 */
	lam_tr_dtypefree(dt);
	if (dt->dt_label > TRDTMAX) {
		free((char *) dt);
	}

	return(MPI_SUCCESS);
}

/*
 *	lam_dtsndrcv
 *
 *	Function:	- copy MPI message from buffer into another
 *			- send/recv done if cannot optimize
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- receive buffer
 *			- receive count
 *			- receive datatype
 *			- tag
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */
int
lam_dtsndrcv(sbuf, scount, sdtype, rbuf, rcount, rdtype, tag, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
int			tag;
MPI_Comm		comm;

{
	int		err;			/* error code */
	int		size;			/* packed size */
	int		rank;			/* caller's rank */
	MPI_Status	stat;			/* status info */

	err = lam_mkerr(MPI_ERR_TRUNCATE, 0);
/*
 * If same datatypes used, just copy.
 */
	if (sdtype == rdtype) {
		if (scount <= rcount) {
			lam_dtcpy((char *) rbuf, (char *) sbuf, rcount,
					rdtype);
			err = MPI_SUCCESS;
		}
	}
/*
 * If receive packed.
 */
	else if (rdtype == MPI_PACKED) {
		MPI_Pack_size(scount, sdtype, MPI_COMM_WORLD, &size);
		if (size <= rcount) {
			if (lam_pack(sbuf, scount, sdtype,
						rbuf, rcount) == size) {
				err = MPI_SUCCESS;
			}
		}
	}
/*
 * If send packed.
 */
	else if (sdtype == MPI_PACKED) {
		MPI_Pack_size(rcount, rdtype, MPI_COMM_WORLD, &size);
		if (size >= scount) {
			if (lam_unpack(sbuf, scount,
					rbuf, rcount, rdtype) == scount) {
				err = MPI_SUCCESS;
			}
		}
	}
/*
 * Let send/recv handle it.
 */
	else {
		MPI_Comm_rank(comm, &rank);
		err = MPI_Sendrecv(sbuf, scount, sdtype, rank, tag,
				rbuf, rcount, rdtype, rank, tag, comm, &stat);
	}

	return(err);
}

/*
 *	lam_dtcpy
 *
 *	Function:	- copy a datatype from a buffer into another
 *			- copying is done recursively
 *	Accepts:	- destination buffer
 *			- source buffer
 *			- count
 *			- datatype
 */
void
lam_dtcpy(destbuf, srcbuf, count, dtype)

char			*destbuf;
char			*srcbuf;
int			count;
MPI_Datatype		dtype;

{
/*
 * Handle the different datatypes.
 */
	switch(dtype->dt_format) {

	case LAM_DTBASIC:
		if ((dtype == MPI_UNSIGNED_CHAR) || (dtype == MPI_PACKED) ||
				(dtype == MPI_BYTE) || (dtype == MPI_CHAR) ||
				(dtype == MPI_F_CHARACTER)) {

			memcpy(destbuf, srcbuf, count);

		}
		else if ((dtype == MPI_INT) || (dtype == MPI_UNSIGNED) ||
				(dtype == MPI_F_LOGICAL) ||
				(dtype == MPI_F_INTEGER)) {

			memcpy(destbuf, srcbuf, count * sizeof(int));
		}
		else if ((dtype == MPI_LONG) ||
				(dtype == MPI_UNSIGNED_LONG)) {

			memcpy(destbuf, srcbuf, count * sizeof(long));
		}
		else if ((dtype == MPI_SHORT) ||
				(dtype == MPI_UNSIGNED_SHORT)) {

			memcpy(destbuf, srcbuf, count * sizeof(short));
		}
		else if ((dtype == MPI_FLOAT) || (dtype == MPI_F_REAL)) {

			memcpy(destbuf, srcbuf, count * sizeof(float));
		}
		else if ((dtype == MPI_DOUBLE) ||
				(dtype == MPI_F_DOUBLE_PRECISION)) {

			memcpy(destbuf, srcbuf, count * sizeof(double));
		}
		else if (dtype == MPI_LONG_DOUBLE) {
#if __STDC__
			memcpy(destbuf, srcbuf, count * sizeof(long double));
#else
			memcpy(destbuf, srcbuf, count * sizeof(double));
#endif
		}
		else if (dtype == MPI_F_COMPLEX) {

			memcpy(destbuf, srcbuf, 2 * count * sizeof(float));
		}
		else if (dtype == MPI_F_DOUBLE_COMPLEX) {

			memcpy(destbuf, srcbuf, 2 * count * sizeof(double));
		}

		break;

	case LAM_DTCONTIG:
		lam_dtcpy(destbuf, srcbuf,
				count * dtype->dt_count, dtype->dt_dtype);
		break;

	case LAM_DTVECTOR:
		cpy_hvector(destbuf, srcbuf, count, dtype, 1);
		break;

	case LAM_DTHVECTOR:
		cpy_hvector(destbuf, srcbuf, count, dtype, 0);
		break;

	case LAM_DTINDEXED:
		cpy_hindexed(destbuf, srcbuf, count, dtype, 1);
		break;

	case LAM_DTHINDEXED:
		cpy_hindexed(destbuf, srcbuf, count, dtype, 0);
		break;

	case LAM_DTSTRUCT:
		cpy_struct(destbuf, srcbuf, count, dtype);
		break;
	}
}

/*
 *	cpy_hvector
 *
 *	Function:	- copy vector or hvector
 */
static void
cpy_hvector(dest, src, num, dtype, fl_vec)

char			*dest;
char			*src;
int			num;
MPI_Datatype		dtype;
int			fl_vec;

{
	int		i, j;			/* favourite indices */
	int		extent;			/* datatype extent */
	int		stride;			/* stride in bytes */
	char		*sbuf, *dbuf;		/* temp. buffers */

	extent = dtype->dt_upper - dtype->dt_lower;
	stride = dtype->dt_stride;
	if (fl_vec) stride *= dtype->dt_dtype->dt_upper -
				dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		sbuf = src;
		dbuf = dest;

		src += extent;
		dest += extent;

		for (j = 0; j < dtype->dt_count; ++j) {

			lam_dtcpy(dbuf, sbuf,
					dtype->dt_length, dtype->dt_dtype);

			sbuf += stride;
			dbuf += stride;
		}
	}
}

/*
 *	cpy_hindexed
 *
 *	Function:	- copy indexed or hindexed
 */
static void
cpy_hindexed(dest, src, num, dtype, fl_idx)

char			*dest;
char			*src;
int			num;
MPI_Datatype		dtype;
int			fl_idx;

{
	int		i, j;			/* favourite indices */
	int		extent;			/* datatype extent */
	int		factor;			/* extent factor */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	char		*sbuf, *dbuf;		/* temp. buffers */

	extent = dtype->dt_upper - dtype->dt_lower;
	factor = (fl_idx == 0) ? 1 :
			dtype->dt_dtype->dt_upper - dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		sbuf = src;
		dbuf = dest;

		src += extent;
		dest += extent;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp) {

			lam_dtcpy(dbuf + (*disp * factor),
					sbuf + (*disp * factor),
					*len, dtype->dt_dtype);
		}
	}
}

/*
 *	cpy_struct
 *
 *	Function:	- copy struct
 */
static void
cpy_struct(dest, src, num, dtype)

char			*dest;
char			*src;
int			num;
MPI_Datatype		dtype;

{
	int		i, j;			/* favourite indices */
	int		extent;			/* datatype extent */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	MPI_Datatype	*type;			/* ptr datatypes */
	char		*sbuf, *dbuf;		/* temp. buffers */

	extent = dtype->dt_upper - dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		sbuf = src;
		dbuf = dest;

		src += extent;
		dest += extent;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;
		type = dtype->dt_dtypes;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp, ++type) {

			lam_dtcpy(dbuf + *disp, sbuf + *disp, *len, *type);
		}
	}
}

/*
 *	lam_dtblock
 *
 *	Function:	- add data block to datatype
 *	Accepts:	- datatype block is being added to
 *			- block datatype
 *			- block count
 *			- block displacement
 */
void
lam_dtblock(new, old, count, disp)

MPI_Datatype		new;
MPI_Datatype		old;
int			count;
int			disp;

{
	int		extent;
	int		upper;
	int		lower;
	int		dataup;
	int		datalow;

	extent = (old->dt_upper - old->dt_lower) * (count - 1);
	
	if (extent > 0) {
		upper = old->dt_upper + extent + disp;
		dataup = old->dt_dataup + extent + disp;
		lower = old->dt_lower + disp;
		datalow = old->dt_datalow + disp;
	} else {
		upper = old->dt_upper + disp;
		dataup = old->dt_dataup + disp;
		lower = old->dt_lower + extent + disp;
		datalow = old->dt_datalow + extent + disp;
	}
/*
 * First block in the new type.
 */
	if (new->dt_size < 0) {
		new->dt_upper = upper;
		new->dt_dataup = dataup;
		new->dt_lower = lower;
		new->dt_datalow = datalow;

		new->dt_size = count * old->dt_size;
		new->dt_nelem = (old->dt_size > 0) ? count : 0;
	}
/*
 * Adding a new block.
 */
	else {
		if (old->dt_size > 0) {
			new->dt_size += count * old->dt_size;
			new->dt_nelem += count;
			
			if (dataup > new->dt_dataup) {
				new->dt_dataup = dataup;
			}
			if (datalow < new->dt_datalow) {
				new->dt_datalow = datalow;
			}
		}
		
		if (old->dt_flags & LAM_DTHASUB) {
			if (new->dt_flags & LAM_DTHASUB) {
				if (upper > new->dt_upper) {
					new->dt_upper = upper;
				}
			} else {
				new->dt_flags |= LAM_DTHASUB;
				new->dt_upper = upper;
			}
		} else {
			if (!(new->dt_flags & LAM_DTHASUB)) {
				new->dt_upper = new->dt_dataup;
			}
		}

		if (old->dt_flags & LAM_DTHASLB) {
			if (new->dt_flags & LAM_DTHASLB) {
				if (lower < new->dt_lower) {
					new->dt_lower = lower;
				}
			} else {
				new->dt_flags |= LAM_DTHASLB;
				new->dt_lower = lower;
			}
		} else {
			if (!(new->dt_flags & LAM_DTHASLB)) {
				new->dt_lower = new->dt_datalow;
			}
		}
	}
}

/*
 *	lam_dtalign
 *
 *	Function:	- adjust extent of datatype to account for alignment
 *	Accepts:	- datatype
 */
void
lam_dtalign(dtype)

MPI_Datatype		dtype;

{
	int		extent;			/* extent of datatype */
	int		epsilon;		/* extra needed for alignment */

	if (! (dtype->dt_flags & LAM_DTHASUB)) {
		
		extent = dtype->dt_upper - dtype->dt_lower;

		if (extent > 0) {
			epsilon = extent % dtype->dt_align;
			if (epsilon != 0) {
				epsilon = dtype->dt_align - epsilon;
			}
		}
		else {
			epsilon = (-extent) % dtype->dt_align;
		}

		dtype->dt_upper += epsilon;
	}
}

/*
 *	lam_dtbuffer
 *
 *	Function:	- allocate buffer for copies of datatype
 *	Accepts:	- datatype
 *			- count of copies
 *			- buffer (out)
 *			- origin of buffer (out)
 *	Returns:	- MPI_SUCCESS or error code
 */
int
lam_dtbuffer(dtype, count, buffer, origin)

MPI_Datatype		dtype;
int			count;
char			**buffer;
char			**origin;

{
	int		extent;			/* space required for extent */
	int		up;			/* upper limit */
	int		low;			/* lower limit */

	if (count <= 0) {
		*buffer = *origin = 0;
		return(MPI_SUCCESS);
	}

	extent = (dtype->dt_upper - dtype->dt_lower) * (count - 1);

	if (extent >= 0) {
		up = dtype->dt_dataup + extent;
		low = dtype->dt_datalow;
	} else {
		up = dtype->dt_dataup;
		low = dtype->dt_datalow + extent;
	}

	if (up > low) {
		*buffer = malloc((unsigned) (up - low));
		if (*buffer == 0) {
			return(lam_mkerr(MPI_ERR_OTHER, errno));
		}
		*origin = *buffer - low;
	} else {
		*buffer = *origin = 0;
	}

	return(MPI_SUCCESS);
}
